"""Keypress."""

import curses.ascii


def is_escape(key):
    """Detect ESC key.

    This is used to exit the application.

    Args:
        key (str): Individual characters are returned as 1-character
            strings, and special keys such as function keys
            return longer strings containing a key name such as
            KEY_UP or ^G.

    Returns:
        bool: Returns `True` if pressed key is ESC key.
        Returns `False` otherwise.
    """
    if isinstance(key, str) and len(key) == 1:
        return ord(key) == curses.ascii.ESC
    return False


def is_ctrl_c(key):
    """Detect Ctrl+c key combination.

    This is used to exit the application.

    Args:
        key (str): Character to check.

    Returns:
        bool: Returns `True` if Ctrl+c is pressed.
        Returns `False` otherwise.
    """
    return key == "\x03"


def is_ctrl_t(key):
    """Detect Ctrl+t key combination.

    This is used to share results on twitter

    Args:
        key (str): Character to check.
    Returns:
        bool: Return `True` if Ctrl+t is pressed.
        Return `False` otherwise.
    """
    return key == "\x14"


def is_ctrl_backspace(key):
    """Detect Ctrl+backspace key combination.

    Used to delete the last typed word

    Args:
        key (str): Character to check.

    Returns:
        bool: Returns `True` if Ctrl+backspace is pressed.
        Returns `False` otherwise.
    """
    return key == "\x17"


def is_backspace(key):
    """Detect BACKSPACE key.

    Args:
        key (str): Character to check.

    Returns:
        bool: Returns `True` if pressed key is BACKSPACE key.
        Returns `False` otherwise.
    """
    if key in ("KEY_BACKSPACE", "\b", "\x7f"):
        return True
    return key in (curses.KEY_BACKSPACE, curses.KEY_DC)


def is_null(key):
    """Detect null keys like super key.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if "null" key or `False` otherwise.
    """
    if isinstance(key, str) and len(key) == 1:
        return ord(key) == 0
    return key == ""


def is_enter(key):
    """Detect enter key.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if line feed or `False` otherwise.
    """
    return key == "\n"


def is_tab(key):
    """Detect tab key to start mitype again.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if tab key or `False` otherwise.
    """
    return key == "\t"


def is_resize(key):
    """Detect if terminal was resized.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if resize request or `False` otherwise.
    """
    return key == "KEY_RESIZE"


def is_ignored_key(key):
    """Detect if key press should be ignored.

    Special function keys, page navigation keys must be ignored.

    Args:
        key (str): Character to check.

    Returns:
        bool: Returns `True` if pressed key must be ignored or `False`
        otherwise.
    """
    return isinstance(key, int)


def is_valid_initial_key(key):
    """Detect if the pressed key is a valid key to start timer.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if key is a valid text character or `False` otherwise.
    """
    return not (
        is_resize(key)
        or is_null(key)
        or is_enter(key)
        or is_escape(key)
        or is_backspace(key)
        or is_tab(key)
        or is_ignored_key(key)
    )


def is_right_arrow_key(key):
    """Detect right arrow key.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if key is a valid text character or `False` otherwise.
    """
    if key == "KEY_RIGHT":
        return True
    return key == curses.KEY_RIGHT


def is_left_arrow_key(key):
    """Detect left arrow key.

    Args:
        key (str): Character to check.

    Returns:
        bool: `True` if key is a valid text character or `False` otherwise.
    """
    if key == "KEY_LEFT":
        return True
    return key == curses.KEY_LEFT



"""Custom handlers for received signals."""
import signal
import sys


def exit_on_signal(signum, frame):
    """Exit when SIGINT signal is received."""
    sys.exit(0)


signal.signal(signal.SIGINT, exit_on_signal)



"""Calculations."""

import math

from mitype import timer


def first_index_at_which_strings_differ(string1, string2):
    """Return index at which there is a change in strings.

    This is used to determine the index up to which text must be dimmed and
    after which must be coloured red (indicating mismatch).

    Args:
        string1 (str): The string which is a combination of
            last typed keys in a session.
        string2 (str): The string corresponding to sample text.

    Returns:
        int: Index at which mismatch occurs for the first time.
    """
    length = min(len(string1), len(string2))

    for index in range(length):
        if string1[index] != string2[index]:
            return index
    return length


def number_of_lines_to_fit_text_in_window(string, window_width):
    """Count number of lines required for displaying text.

    Args:
        string (str): String containing sample text.
        window_width (int): Width of terminal.

    Returns:
        int: The number of lines required to display sample text
    """
    return int(math.ceil(len(string) / window_width))


def get_space_count_after_ith_word(index, text):
    """Return number of spaces after a given word.

    Args:
        index (int): Index of word in text list
        text(str): Text without appending extra spaces

    Returns:
        int: The number of spaces required after ith word
    """
    count = 0
    while index < len(text) and text[index] == " ":
        index += 1
        count += 1
    return count


def word_wrap(text, width):
    """Wrap text on the screen according to the window width.

    Returns text with extra spaces which makes the string word wrap.

    Args:
        text (str): Text to wrap.
        width (int): Width to wrap around.

    Returns:
        str: Return altered text.
    """
    # For the end of each line, move backwards until you find a space.
    # When you do, append those many spaces after the single space.
    for line in range(
        1,
        number_of_lines_to_fit_text_in_window(text, width) + 1,
    ):
        # Current line fits in the window
        if line * width >= len(text):
            continue

        # Last cell of that line
        index = line * width - 1

        # Continue if already a space
        if text[index] == " ":
            continue

        # Find last occurrence of space on that line
        index = text[:index].rfind(" ")

        space_count = line * width - index
        space_string = " " * space_count

        text = text[:index] + space_string + text[index + 1 :]
    return text


def speed_in_wpm(text, start_time):
    """Calculate typing speed in WPM.

    Args:
        text (list): List of words from sample text.
        start_time (float): The time when user starts typing
            the sample text.

    Returns:
        str: Speed in WPM up to 2 decimal places.
    """
    time_taken = timer.get_elapsed_minutes_since_first_keypress(start_time)
    wpm = len(text) / time_taken

    return f"{wpm:.2f}"


def accuracy(total_chars_typed, wrongly_typed):
    """Get accuracy for the current test.

    Args:
        total_chars_typed (int): Total characters typed.
        wrongly_typed (int): Mistyped characters.

    Returns:
        float: Return accuracy.
    """
    return ((total_chars_typed - wrongly_typed) / total_chars_typed) * 100



"""
Start of app.

Parses command line arguments and decides and fills text accordingly.
"""

import argparse
import os
import random
import sys

import mitype
import mitype.database
from mitype.history import show_history


def resolve_commandline_arguments():
    """Parse CLI arguments and return practice text details.

    Returns:
        (str, Union[str, int]): Tuple of text content and text ID.
    """
    opt = parse_arguments()
    if opt.version:
        display_version()
        sys.exit(0)

    elif opt.history is not None:
        show_history(opt.history)
        sys.exit(0)

    elif opt.file:
        text, text_id = load_text_from_file(opt.file)

    elif opt.id:
        text, text_id = load_from_database(opt.id)

    elif opt.difficulty:
        text, text_id = load_based_on_difficulty(opt.difficulty)

    else:
        text, text_id = load_based_on_difficulty()

    return text, text_id


def parse_arguments():
    """Parse command line arguments.

    Returns:
        str: Parsed command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Process mitype command line arguments",
    )

    parser.add_argument(
        "-V",
        "--version",
        default=False,
        action="store_true",
        help="Show mitype version",
    )

    parser.add_argument(
        "-f",
        "--file",
        metavar="FILENAME",
        default=None,
        type=str,
        help="File to use text from as sample text",
    )

    parser.add_argument(
        "-i",
        "--id",
        metavar="id",
        default=None,
        type=int,
        help="ID to retrieve text from database",
    )

    parser.add_argument(
        "-d",
        "--difficulty",
        metavar="N",
        default=2,
        type=int,
        help="Choose difficulty within range 1-5",
    )

    parser.add_argument(
        "-H",
        "--history",
        nargs="?",
        default=None,
        const=-1,
        type=int,
        help="Show mitype score history",
    )

    return parser.parse_args()


def display_version():
    """Display version."""
    print(f"Mitype version {mitype.__version__}")


def load_text_from_file(file_path):
    """Load file contents.

    Args:
        file_path (str): Full path to text to load.

    Returns:
        (str, str): Tuple of text content followed by file path.
    """
    if os.path.isfile(file_path):
        with open(file_path, encoding="utf-8") as file:
            text = file.read()
            filename = os.path.basename(file_path)
        return text, filename

    print("Cannot open file -", file_path)
    sys.exit(0)


def load_from_database(text_id):
    """Load given text from database with given id.

    Args:
        text_id (int): Row identifier of database text to load.

    Returns:
        (str, int): Tuple of text content followed by DB row identifier.
    """
    row_count = 6000
    if 1 <= text_id <= row_count:
        text = mitype.database.fetch_text_from_id(text_id)
        return text, text_id

    print("ID must be in range [1,6000]")
    sys.exit(1)


def load_based_on_difficulty(difficulty_level=random.randrange(1, 6)):
    """
    Load text of given difficulty from database if parameter is passed.

    Defaults to random difficulty level when none is provided.

    Args:
        difficulty_level (int): difficulty level in a range of 1 - 5

    Returns:
        (str, int): Tuple of text content followed by DB row identifier.
    """
    max_level = 5

    if 1 <= difficulty_level <= max_level:
        # Each difficulty section has 6000/5 = 1200 texts each
        upper_limit = difficulty_level * 1200
        lower_limit = upper_limit - 1200 + 1

        text_id = random.randrange(lower_limit, upper_limit + 1)
        text = mitype.database.fetch_text_from_id(text_id)

        return text, text_id

    print("Select a difficulty level in range [1,5]")
    sys.exit(1)



"""Handle with user history."""

import csv
import os
import time
from datetime import date


def history_file_absolute_path():
    """Get full path of history file.

    This is required for later reading or modifying records from history file.

    Returns:
        str: The path of history file.
    """
    history_filename = ".mitype_history.csv"

    return os.path.join(os.path.expanduser("~"), history_filename)


def get_history_records(number_of_records):
    """Get records from history.

    Defaults to -1 if argument value not provided on command line.

    Args:
        number_of_records (int): Number of last records to print.
    Returns:
        list: A list of records. The len of this list is `number_of_records` or all records
    """
    history_file_path = history_file_absolute_path()

    if not os.path.exists(history_file_path):
        return []

    with open(history_file_path, encoding="utf-8") as file:
        history_reader = csv.reader(file)

        try:
            # Skip csv header
            next(history_reader)
        except StopIteration:
            # No header found on the file, meaning the file is empty
            return []

        data = list(history_reader)
        total_records = len(data)

        if number_of_records <= -1 or number_of_records >= total_records:
            number_of_records = total_records

        start_count = 0
        if number_of_records < total_records and number_of_records != -1:
            start_count = total_records - number_of_records
        return data[start_count:total_records]


def show_history(number_of_records):
    """Show records from history.

    Defaults to -1 if argument value not provided on command line.

    Args:
        number_of_records (int): Number of last records to print.
    """
    records = get_history_records(number_of_records)

    if len(records) == 0:
        print("0 records found")

    print("Last {} records:".format(len(records)))
    print("ID\tWPM\tDATE\t\tTIME\t\tACCURACY")
    for record in records:
        formatted_row_data = "\t".join(record)
        print(formatted_row_data + "%")


def save_history(text_id, current_speed_wpm, accuracy):
    """Save test stats to history file.

    Args:
        text_id (int): Row identifier of database text loaded.
        current_speed_wpm (float): Speed result from test.
        accuracy (str): Accuracy result from test.
    """
    history_path = history_file_absolute_path()

    file_exists = os.path.isfile(history_path)

    with open(history_path, mode="a", newline="", encoding="utf-8") as history:
        csv_history = csv.writer(history)

        if not file_exists:
            row = ["ID", "WPM", "DATE", "TIME", "ACCURACY"]
            csv_history.writerow(row)

        current_time = time.strftime("%H:%M:%S", time.localtime())

        test_data = [
            text_id,
            current_speed_wpm,
            date.today(),
            current_time,
            accuracy,
        ]
        csv_history.writerow(test_data)



"""Timer."""

import time


def get_elapsed_minutes_since_first_keypress(start_time):
    """Get time elapsed since initial keypress.

    This is required to calculate speed.

    Args:
        start_time (float): The time when user starts typing
            the sample text.

    Returns:
        float: Time elapsed since start of typing session till calling
        this function.
    """
    return (time.time() - start_time) / 60



"""Setup script for mitype."""
from setuptools import setup

import versioneer

if __name__ == "__main__":
    setup(
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        download_url="https://github.com/mithil467/mitype/archive/v%s.tar.gz"
        % versioneer.get_version(),
    )



"""Deals with fetching texts from database."""

import os
import sqlite3


def database_file_absolute_path():
    """Get full path of directory where source files are stored.

    This is required for later fetching entry from data.db which is
    stored in same directory as app.

    Returns:
        str: The path of directory of source file.
    """
    database_filename = "data.db"
    database_directory_absolute_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(
        database_directory_absolute_path,
        database_filename,
    )


def fetch_text_from_id(serial_id):
    """Fetch row from data.db database.

    Args:
        serial_id (int): The unique ID of database entry.

    Returns:
        str: The text corresponding to the entry_id.
    """
    database_file = database_file_absolute_path()
    connection = sqlite3.connect(database_file)
    cursor = connection.cursor()

    # For details related to the database schema check CONTRIBUTING.md
    cursor.execute("SELECT txt FROM data where id=?", (serial_id,))
    text = cursor.fetchone()[0]
    connection.close()

    return text



"""This is the Mitype main app script."""

import curses
import os
import sys
import time
import webbrowser

import mitype.signals
from mitype.calculations import (
    accuracy,
    first_index_at_which_strings_differ,
    get_space_count_after_ith_word,
    number_of_lines_to_fit_text_in_window,
    speed_in_wpm,
    word_wrap,
)
from mitype.commandline import load_from_database, resolve_commandline_arguments
from mitype.history import save_history
from mitype.keycheck import (
    is_backspace,
    is_ctrl_backspace,
    is_ctrl_c,
    is_ctrl_t,
    is_enter,
    is_escape,
    is_left_arrow_key,
    is_resize,
    is_right_arrow_key,
    is_tab,
    is_valid_initial_key,
)
from mitype.timer import get_elapsed_minutes_since_first_keypress


class App:
    """Class for enclosing all methods required to run Mitype."""

    def __init__(self):
        """Initialize the application class."""
        # Start the parser
        self.text, self.text_id = resolve_commandline_arguments()
        self.tokens = self.text.split()

        # Squash multiple spaces, tabs, newlines to single space
        self.text = " ".join(self.tokens)
        self.text_backup = self.text

        # Current typed word and entire string
        self.current_word = ""
        self.current_string = ""

        self.key = ""
        # First valid key press
        self.first_key_pressed = False
        # Stores keypress, time tuple
        self.key_strokes = []

        self.mistyped_keys = []

        # Time at which test started
        self.start_time = 0
        # Time at which test ended
        self.end_time = 0

        # Keep track of the token index in text
        self.token_index = 0
        # mode = 0 when in test
        # mode = 1 when in replay
        self.mode = 0

        self.window_height = 0
        self.window_width = 0

        self.number_of_lines_to_print_text = 0

        # Restrict current word length to a limit
        # Used to highlight once the limit is reached
        self.current_word_limit = 25

        self.test_complete = False

        # Real-time speed, the value at the end of the test is the result
        # And a few other stats
        self.current_speed_wpm = 0
        self.accuracy = 0
        self.time_taken = 0

        self.total_chars_typed = 0
        self.text_without_spaces = self.text_backup.replace(" ", "")

        # Color mapping
        self.Color = None

        sys.stdout = sys.__stdout__

        # Set ESC delay to 0 (default 1 on linux)
        os.environ.setdefault("ESCDELAY", "0")

        # Start curses on main
        curses.wrapper(self.main)

    def main(self, win):
        """Respond to user inputs.

        This is where the infinite loop is executed to continuously serve events.

        Args:
            win (any): Curses window object.
        """
        # Initialize windows
        self.initialize(win)

        while True:
            # Typing mode
            key = self.keyinput(win)

            if not self.first_key_pressed:
                if is_escape(key) or is_ctrl_c(key):
                    sys.exit(0)

                if is_left_arrow_key(key):
                    self.switch_text(win, -1)

                if is_right_arrow_key(key):
                    self.switch_text(win, 1)

            # Test mode
            if self.mode == 0:
                self.typing_mode(win, key)

            # Again mode
            else:
                # Tab to retry last test
                if is_tab(key):
                    win.clear()
                    self.reset_test()
                    self.setup_print(win)
                    self.update_state(win)

                # Replay
                if is_enter(key):
                    self.replay(win)

                # Tweet result
                if is_ctrl_t(key):
                    self.share_result()

            # Refresh for changes to show up on window
            win.refresh()

    def initialize(self, win):
        """Configure the initial state of the curses interface.

        Args:
            win (any): Curses window.
        """
        self.window_height, self.window_width = self.get_dimensions(win)

        # This works by adding extra spaces to the text where needed
        self.text = word_wrap(self.text, self.window_width)

        # Check if we can fit text in current window after adding word wrap
        self.screen_size_check()

        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_GREEN)
        curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_RED)
        curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_BLUE)
        curses.init_pair(4, curses.COLOR_WHITE, curses.COLOR_YELLOW)
        curses.init_pair(5, curses.COLOR_WHITE, curses.COLOR_CYAN)
        curses.init_pair(6, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
        curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_WHITE)

        class Color:
            """Color mapping."""

            GREEN = curses.color_pair(1)
            RED = curses.color_pair(2)
            BLUE = curses.color_pair(3)
            YELLOW = curses.color_pair(4)
            CYAN = curses.color_pair(5)
            MAGENTA = curses.color_pair(6)
            BLACK = curses.color_pair(7)

        self.Color = Color

        # This sets input to be a non-blocking call and will block for 100ms
        # Returns -1 if no input found at the end of time
        win.nodelay(True)
        win.timeout(100)

        self.setup_print(win)

    def setup_print(self, win):
        """Print setup text at beginning of each typing session.

        Args:
            win (any): Curses window object.
        """
        win.addstr(0, 0, f" ID:{self.text_id} ", self.Color.CYAN)
        win.addstr(0, self.window_width // 2 - 4, " MITYPE ", self.Color.BLUE)

        # Text is printed BOLD initially
        # It is dimmed as user types on top of it
        win.addstr(2, 0, self.text, curses.A_BOLD)

        self.print_realtime_wpm(win)

        # Set cursor position to beginning of text
        win.move(2, 0)

    def clear_line(self, win, line):
        """Clear a line on the window.

        Args:
            win (any): Curses window.
            line (int): Line number.
        """
        # Cursor advances to next cell after the character is printed
        # This causes scroll with addstr on the last line which is disabled
        # Hence using insstr instead
        win.insstr(line, 0, " " * self.window_width)

    def update_state(self, win):
        """Report on typing session results.

        Args:
            win (any): Curses window.
        """
        self.clear_line(win, self.number_of_lines_to_print_text)
        self.clear_line(win, self.number_of_lines_to_print_text + 2)
        self.clear_line(win, self.number_of_lines_to_print_text + 4)

        # Highlight in RED if word reaches the word limit length
        if len(self.current_word) >= self.current_word_limit:
            win.addstr(
                self.number_of_lines_to_print_text,
                0,
                self.current_word,
                self.Color.RED,
            )
        else:
            win.addstr(self.number_of_lines_to_print_text, 0, self.current_word)

        # Text is printed BOLD initially
        # It is dimmed as user types on top of it
        win.addstr(2, 0, self.text, curses.A_BOLD)
        win.addstr(2, 0, self.text[0 : len(self.current_string)], curses.A_DIM)

        index = first_index_at_which_strings_differ(self.current_string, self.text)
        # Check if difference was found
        if index < len(self.current_string) <= len(self.text):
            self.mistyped_keys.append(len(self.current_string) - 1)

        win.addstr(
            2 + index // self.window_width,
            index % self.window_width,
            self.text[index : len(self.current_string)],
            self.Color.RED,
        )

        # End of test, all characters are typed out
        if index == len(self.text):
            self.test_end(win)

        win.refresh()

    def test_end(self, win):
        """Trigger at the end of the test.

        Display options for the user to choose at the end of the test.
        Display stats.

        Args:
            win (any): Curses window.
        """
        # Highlight mistyped characters
        for i in self.mistyped_keys:
            win.addstr(
                2 + i // self.window_width,
                i % self.window_width,
                self.text[i],
                self.Color.RED,
            )

        curses.curs_set(0)

        # Calculate stats at the end of the test
        if self.mode == 0:
            self.current_speed_wpm = speed_in_wpm(self.tokens, self.start_time)
            total_chars_in_text = len(self.text_without_spaces)
            wrongly_typed_chars = self.total_chars_typed - total_chars_in_text
            self.accuracy = accuracy(self.total_chars_typed, wrongly_typed_chars)
            self.time_taken = get_elapsed_minutes_since_first_keypress(self.start_time)

            self.mode = 1
            # Find time difference between the key strokes
            # The key_strokes list is storing the time at which the key is pressed
            for index in range(len(self.key_strokes) - 1, 0, -1):
                self.key_strokes[index][0] -= self.key_strokes[index - 1][0]

            self.key_strokes[0][0] = 0

        win.addstr(self.number_of_lines_to_print_text, 0, " Your typing speed is ")
        win.addstr(" " + self.current_speed_wpm + " ", self.Color.MAGENTA)
        win.addstr(" WPM ")

        win.addstr(
            self.number_of_lines_to_print_text + 2,
            1,
            " Enter ",
            self.Color.BLACK,
        )
        win.addstr(" to see replay, ")

        win.addstr(" Tab ", self.Color.BLACK)
        win.addstr(" to retry.")

        win.addstr(
            self.number_of_lines_to_print_text + 3,
            1,
            " Arrow keys ",
            self.Color.BLACK,
        )
        win.addstr(" to change text.")

        win.addstr(
            self.number_of_lines_to_print_text + 4,
            1,
            " CTRL+T ",
            self.Color.BLACK,
        )
        win.addstr(" to tweet result.")

        self.print_stats(win)

        self.first_key_pressed = False
        self.end_time = time.time()
        self.current_string = ""
        self.current_word = ""
        self.token_index = 0

        self.start_time = 0
        if not self.test_complete:
            win.refresh()
            save_history(self.text_id, self.current_speed_wpm, f"{self.accuracy:.2f}")
            self.test_complete = True

    def typing_mode(self, win, key):
        """Start recording typing session progress.

        Args:
            win (any): Curses window.
            key (str): First typed character of the session.
        """
        # Note start time when first valid key is pressed
        if not self.first_key_pressed and is_valid_initial_key(key):
            self.start_time = time.time()
            self.first_key_pressed = True

        if is_resize(key):
            self.resize(win)

        if not self.first_key_pressed:
            return

        self.key_strokes.append([time.time(), key])

        self.print_realtime_wpm(win)

        self.key_printer(win, key)

    @staticmethod
    def keyinput(win):
        """Retrieve next character of text input.

        Args:
            win (any): Curses window.

        Returns:
            str: Value of typed key.
        """
        key = ""
        try:
            key = win.get_wch()
            if isinstance(key, int):
                if key in (curses.KEY_BACKSPACE, curses.KEY_DC):
                    return "KEY_BACKSPACE"
                if key == curses.KEY_RESIZE:
                    return "KEY_RESIZE"
            return key
        except curses.error:
            return ""

    def key_printer(self, win, key):
        """Print required key to terminal.

        Args:
            win (any): Curses window object.
            key (str): Individual characters are returned as 1-character
                strings, and special keys such as function keys
                return longer strings containing a key name such as
                KEY_UP or ^G.
        """
        # Reset test
        if is_escape(key):
            self.reset_test()

        elif is_ctrl_c(key):
            sys.exit(0)

        # Handle resizing
        elif is_resize(key):
            self.resize(win)

        # Check for backspace
        elif is_backspace(key):
            self.erase_key()

        elif is_ctrl_backspace(key):
            self.erase_word()

        # Ignore spaces at the start of the word (Plover support)
        elif key == " " and len(self.current_word) < self.current_word_limit:
            if self.current_word != "":
                self.check_word()

        elif is_valid_initial_key(key):
            self.appendkey(key)
            self.total_chars_typed += 1

        # Update state of window
        self.update_state(win)

    def resize(self, win):
        """Respond to window resize events.

        Args:
            win (any): Curses window.
        """
        win.clear()

        self.window_height, self.window_width = self.get_dimensions(win)
        self.text = word_wrap(self.text_backup, self.window_width)

        self.screen_size_check()

        self.print_realtime_wpm(win)
        self.setup_print(win)
        self.update_state(win)

    def print_stats(self, win):
        """Print the bottom stats bar after each run.

        Args:
            win (any): Curses window.
        """
        win.addstr(
            self.window_height - 1,
            0,
            f" WPM: {self.current_speed_wpm} ",
            self.Color.MAGENTA,
        )

        win.addstr(
            f" Time: {self.time_taken*60:.2f}s ",
            self.Color.GREEN,
        )

        win.addstr(
            f" Accuracy: {self.accuracy:.2f}% ",
            self.Color.CYAN,
        )

    def print_realtime_wpm(self, win):
        """Print realtime wpm during the test.

        Args:
            win (any): Curses window.
        """
        current_wpm = 0
        total_time = mitype.timer.get_elapsed_minutes_since_first_keypress(
            self.start_time,
        )
        if total_time != 0:
            words = self.current_string.split()
            word_count = len(words)
            current_wpm = word_count / total_time

        win.addstr(
            0,
            self.window_width - 14,
            f" {current_wpm:.2f} ",
            self.Color.CYAN,
        )
        win.addstr(" WPM ")

    def replay(self, win):
        """Play out a recording of the user's last session.

        Args:
            win (any): Curses window.
        """
        win.clear()
        self.print_stats(win)
        win.addstr(self.number_of_lines_to_print_text + 2, 0, " " * self.window_width)
        curses.curs_set(1)

        win.addstr(
            0,
            self.window_width - 14,
            " " + str(self.current_speed_wpm) + " ",
            self.Color.CYAN,
        )
        win.addstr(" WPM ")

        self.setup_print(win)

        win.timeout(10)
        for key in self.key_strokes:
            time.sleep(key[0])
            _key = self.keyinput(win)
            if is_escape(_key) or is_ctrl_c(_key):
                sys.exit(0)
            self.key_printer(win, key[1])
        win.timeout(100)

    def share_result(self):
        """Open a twitter intent on a browser."""
        message = (
            f"My typing speed is {self.current_speed_wpm} WPM!"
            "Know yours on mitype.\n"
            "https://pypi.org/project/mitype/ by @MithilPoojary\n"
            "#TypingTest"
        )

        # URL encode message
        message = message.replace("\n", "%0D").replace("#", "%23")
        url = "https://twitter.com/intent/tweet?text=" + message
        webbrowser.open(url, new=2)

    def reset_test(self):
        """Reset the data for current typing session."""
        self.mode = 0
        self.current_word = ""
        self.current_string = ""
        self.first_key_pressed = False
        self.key_strokes = []
        self.mistyped_keys = []
        self.start_time = 0
        self.token_index = 0
        self.current_speed_wpm = 0
        self.total_chars_typed = 0
        self.accuracy = 0
        self.time_taken = 0
        curses.curs_set(1)

    def switch_text(self, win, value):
        """Load next or previous text snippet from database.

        Args:
            win (any): Curses window.
            value (int): value to increase or decrement the text ID by.
        """
        if isinstance(self.text_id, str):
            return

        win.clear()

        self.text_id += value
        self.text = load_from_database(self.text_id)[0]
        self.tokens = self.text.split()

        self.text = " ".join(self.tokens)
        self.text_backup = self.text

        self.reset_test()
        self.setup_print(win)
        self.update_state(win)

    @staticmethod
    def get_dimensions(win):
        """Get the height and width of terminal.

        Args:
            win (any): Curses window object.

        Returns:
            (int, int): Tuple of height and width of terminal window.
        """
        return win.getmaxyx()

    def screen_size_check(self):
        """Check if screen size is enough to print text."""
        self.number_of_lines_to_print_text = (
            number_of_lines_to_fit_text_in_window(self.text, self.window_width) + 3
        )
        if self.number_of_lines_to_print_text + 7 >= self.window_height:
            curses.endwin()
            sys.stdout.write("Window too small to print given text")
            sys.exit(1)

    def appendkey(self, key):
        """Append a character to the end of the current word.

        Args:
            key (key): Character to append.
        """
        if len(self.current_word) < self.current_word_limit:
            self.current_word += key
            self.current_string += key

    def erase_key(self):
        """Erase the last typed character."""
        if len(self.current_word) > 0:
            self.current_word = self.current_word[:-1]
            self.current_string = self.current_string[:-1]

    def erase_word(self):
        """Erase the last typed word."""
        if len(self.current_word) > 0:
            index_word = self.current_word.rfind(" ")
            if index_word == -1:
                # Single word.
                word_length = len(self.current_speed_wpm)
                self.current_string = self.current_string[:-word_length]
                self.current_word = ""
            else:
                diff = len(self.current_word) - index_word
                self.current_word = self.current_word[:-diff]
                self.current_string = self.current_string[:-diff]

    def check_word(self):
        """Accept finalized word."""
        spc = get_space_count_after_ith_word(len(self.current_string), self.text)
        if self.current_word == self.tokens[self.token_index]:
            self.token_index += 1
            self.current_word = ""
            self.current_string += spc * " "
        else:
            self.current_word += " "
            self.current_string += " "
